#define KEY_SIZE	4096
#define S_BUF_SIZE	4096
#define C_BUF_SIZE	4096

#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <stdarg.h>
#include "config.h"

#ifdef TLS
#include "tlsutil.h"
#define READ     tls_read
#define WRITE    tls_write
#define SEND     tls_send
#define FGETC    tls_fgetc
#define FPUTS    tls_fputs
#define PUTC     tls_fputc
#define FFLUSH   tls_fflush
#define CLOSE    tls_close
#define FCLOSE   tls_fclose
#define VFPRINTF tls_vfprintf
#define FPRINTF  tls_fprintf
#define ENC	 "TLS"
#define ENC2	 "TLS"
#else
#define READ     read
#define WRITE    write 
#define SEND     send 
#define FGETC    fgetc
#define FPUTS    fputs  
#define FFLUSH   fflush
#define PUTC     putc
#define CLOSE    close
#define FCLOSE   fclose
#define VFPRINTF vfprintf
#define FPRINTF  fprintf
#define ENC	 "RAW"
#define ENC2	 "KEY"
#endif

char	*KEY,
	*TARGET_HOST[ MAXHOSTNAMELEN ],
	*KEY_FILE,
	*ident;
int	key_pos,
	KEY_LEN,
	ENCRYPT_MODE = -1,
	TARGET_HOSTS = 0,
	CURRENT_TARGET = 0,
	READ_IDENT = -1,
	SEND_IDENT = -1,
	LISTEN_PORT = -1,
	TARGET_PORT = -1;
FILE	*ttyout;

extern	char	*rfc931();
extern	int	errno;



/* void encrypt( char *s, int l ) {
*
* Encrypts string using KEY, lame encryption but better than nothing
*
*/
void 
encrypt( char *s, int l ) {
 int	n;

 for ( n = 0 ; n < l ; n++ ) {
	 s[n] += KEY[key_pos];
	 if ( key_pos < KEY_LEN - 1 ) key_pos++; else key_pos = 0;
	}
}




/* void decrypt( char *s, int l )
*
* Decrypts string using KEY
*
*/
void
decrypt( char *s, int l ) {
 int	n;

 for ( n = 0 ; n < l ; n++ ) {
	 s[n] -= KEY[key_pos];
	 if ( key_pos < KEY_LEN - 1 ) key_pos++; else key_pos = 0;
	}
}




/*
* void transmit_loop( int client, int server, char *ident, char *remote_ip, char *remote_host ) 
*
* Transfers ftp control data from client to server and server to client
* and encrypts it, if ENCRYPT_MODE is set to 1 or 2.
*
*/
void 
transmit_loop( int client, int server, char *ident, char *remote_ip, char *remote_host, char *local_host ) {
 fd_set		read_ds,
		write_ds;
 int		n = 0, l, m,
        	client_buf = 0,
		server_buf = 0,
		encrypted = 0;
 char		s_buf[ S_BUF_SIZE ],
		c_buf[ C_BUF_SIZE ],
		*s;

 if (READ_IDENT == 1 && SEND_IDENT == 1) {
	s = malloc(n = (strlen(ident) + strlen(remote_ip) + strlen(remote_host) + 10));
	n = sprintf(s, "IDNT %s@%s:%s\r\n", ident, remote_ip, remote_host);
	write(server, s, n);
	free(s);
 	} else if ( READ_IDENT == 2 ) {
	s = malloc(4096);
	n = read(client, s, 4096);
	if (SEND_IDENT == 1) write(server, s, n);
	free(s);
	}

#ifdef TLS
 server_buf = read( server, s_buf, S_BUF_SIZE );
 write( client, s_buf, server_buf );
 client_buf = read( client, c_buf, C_BUF_SIZE );
 if ( strncmp( c_buf, "AUTH TLS", 8 ) ) {
	 tls_hostname = remote_host;
	 if ( tls_init() ) {
		 return;
		}
	 write( server, "AUTH TLS\r\n", 10 );
	 read( server, s_buf, S_BUF_SIZE );
	 if ( ! strncasecmp("234 AUTH TLS successful", s_buf, 23) ) {
		 if ( tls_connect_ctrl( server )) {
			 return;
			}
		 encrypted = 1;
		}
	} else {
	 encrypted = 2;
	}
 WRITE( server, c_buf, client_buf );
 client_buf = server_buf = 0;
#endif

 while ( 1 ) {
	 FD_ZERO( &read_ds );

	 FD_SET(server, &read_ds);
	 FD_SET(client, &read_ds);

	 select(1 + (server > client ? server : client), &read_ds, 0, 0, 0);
  
	 if ( FD_ISSET(server, &read_ds) ) {
		 n = READ( server, s_buf, S_BUF_SIZE );
		 if ( n <= 0 ) break;

		 if ( ! encrypted ) {
			 switch ( ENCRYPT_MODE ) {
				 case 1: decrypt( s_buf, n ); break;
				 case 2: encrypt( s_buf, n ); break;
				}
			}
		 m = WRITE( client, s_buf, n );
		 if ( n <= 0 || n != m ) break;
		}

	 if ( FD_ISSET(client, &read_ds) ) {
		 n = READ( client, c_buf, C_BUF_SIZE );
		 if ( n <= 0 ) break;

		 if ( ! encrypted ) {
			 switch ( ENCRYPT_MODE ) {
				 case 1: encrypt( c_buf, n ); break;
				 case 2: decrypt( c_buf, n ); break;
				}  
			}
		 m = WRITE( server, c_buf, n);
		 if ( n <= 0 || n != m ) break;
		}
 	}
}





/*
* void del_spaces( char *s )
*
* Deletes all spaces from string s
*
*/
void
del_spaces( char *s ) {
 int	n = 0,
	m = 0;

 while ( s[n] ) {
	 s[m] = s[n];
	 if ( s[n] != ' ' ) m++;
	 n++;
	}
 s[m] = 0;
}





/*
* void read_config( char *f )
*
* Reads dynamic variables from config file and 
* allocates memory for them
*
*/
void
read_config( char *f ) {
 int	n, l;
 long	f_size;
 char	*b,
	*line;
 struct	stat	filestat;

 n = open( f, O_RDONLY );
 fstat( n, &filestat );
 f_size = filestat.st_size;
 b = malloc( f_size );
 read( n, b, f_size );
 close( n );

 line = b; 

 for ( n = 0 ; n < f_size ; n++ ) if ( b[n] == 0 || b[n] == '\n' ) {
	 b[n] = 0;
	 del_spaces( line );
	 if ( ! strncasecmp( line, "TARGET_HOST=", 12 ) ) {
		 if ( ( l = strlen( line + 12 ) ) > 0 ) {
			 TARGET_HOST[TARGET_HOSTS] = malloc( l + 1 );
			 strcpy( TARGET_HOST[TARGET_HOSTS], line + 12 );
			 TARGET_HOSTS++;
			}
		} else if ( ! strncasecmp( line, "TARGET_PORT=", 12 ) ) {
		 if ( atoi( line + 12 ) > 0 ) TARGET_PORT = atoi( line + 12 );
		} else if ( ! strncasecmp( line, "LISTEN_PORT=", 12 ) ) {
		 if ( atoi( line + 12 ) > 0 ) LISTEN_PORT = atoi( line + 12 );
		} else if ( ! strncasecmp( line, "ENCRYPT_MODE=", 13 ) ) {
		 if ( atoi( line + 13 ) >= 0 ) ENCRYPT_MODE = atoi( line + 13 );
		} else if ( ! strncasecmp( line, "KEY_FILE=", 9 ) ) {
		 if ( ( l = open( line + 9, O_RDONLY ) ) > 0 ) {
			 fstat( l, &filestat );
			 KEY = malloc( KEY_LEN = filestat.st_size );
			 read( l, KEY, KEY_LEN );
			 close( l );
			 KEY_FILE = malloc( strlen( line + 9 ) + 1 );
			 strcpy( KEY_FILE, line + 9 );
			}
		} else if ( ! strncasecmp( line, "READ_IDENT=", 11 ) ) {
		 if ( atoi( line + 11 ) >= 0 ) READ_IDENT = atoi( line + 13 );
		} else if ( ! strncasecmp( line, "SEND_IDENT=", 11 ) ) {
		 if ( atoi( line + 11 ) >= 0 ) SEND_IDENT = atoi( line + 13 );
		}
	 line = b + n + 1;
	}
 free( b );
}




/*
* void generate_key( char *f ) 
*
* Generates somewhat random KEY_SIZE byte key 
* and stores it to file named f.
*
*/
void
generate_key( char *f ) {
 FILE	*file;
 int	n = 0;
 char	c;
 struct	timeval tv;

 gettimeofday(&tv,NULL);
 srandom(tv.tv_usec);

 if ( ( file = fopen( f, "r" ) ) == NULL ) {
	 umask(0077 & 0777);
	 if ( ( file = fopen( f, "w+" ) ) != NULL ) {
		 printf( "Creating key-file for dBnc:\n" );
		 for ( n = 0 ; n < KEY_SIZE ; n++ ) {
			 if ( ((int)(100.0*rand()/(RAND_MAX+1.0))) == 25 ) {
				 printf( "x" );
				 sleep( 1 );
				}
			 c = (char)(255.0*rand()/(RAND_MAX+1.0));
			 fwrite( &c, 1, 1, file );
			 printf( "." );
			}
		 printf( "done\n" );
		 exit(0);
		} else {
		 fprintf( stderr, "ERROR: Could not create file!\n" );
		 exit( 2 );
		}
	} else {
	 fprintf( stderr, "ERROR: File already exists!\n" );
	 exit( 2 );
	}
 fclose( file );
}





main ( int argc, char **argv ) {
 struct	sockaddr_in	local_server_addr,
			remote_client_addr,
			remote_server_addr;
 struct hostent		*host_pointer;

 int	server_socket,
	remote_server_socket,
	remote_fd,
	server_fd,
	n,
	show_help = 1;

 char	remote_client[ MAXHOSTNAMELEN ],
	local_host[ MAXHOSTNAMELEN ],
	*ident_p;

 if ( argc == 3 && ( argv[1][0] == '/' || argv[1][0] == '-' ) && argv[1][2] == '\0' ) {
	 switch ( argv[1][1] ) {
		 case 'r':
			 read_config( argv[2] );
			 printf("\n\t. dBnc .......... (c) 2001 .\n");
			 printf("\t:                          :\n");
			 printf("\t:            To            :\n");
			 for ( n = 0 ; n < TARGET_HOSTS ; n++ ) printf("\t:  Host[%i] %15.15s :\n", n + 1, TARGET_HOST[n]);
			 printf("\t:  Port  %17i :\n", TARGET_PORT);
			 if ( ENCRYPT_MODE == 1 ) 
				 printf("\t:  Mode  %17s :\n", ENC2);
				else
				 printf("\t:  Mode  %17s :\n", ENC);
			 if ( SEND_IDENT == 1 )
				 printf("\t:  Ident %17s :\n", "IDNT");
				else
				 printf("\t:  Ident %17s :\n", "NONE");
			 printf("\t:                          :\n");
			 printf("\t:           From           :\n");
			 printf("\t:  Port  %17i :\n", LISTEN_PORT);
			 if ( ENCRYPT_MODE == 2 )
				 printf("\t:  Mode  %17s :\n", ENC2);
				else
				 printf("\t:  Mode  %17s :\n", ENC);
			 if ( READ_IDENT == 1 )
				 printf("\t:  Ident %17s :\n", "IDENTD");
				else if ( READ_IDENT == 2 )
				 printf("\t:  Ident %17s :\n", "IDNT");
				else
				 printf("\t:  Ident %17s :\n", "NONE");
			 printf("\t:                          :\n");
			 printf("\t:   Pid  %17i :\n", getpid());
			 printf("\t:                          :\n");
			 printf("\t:..........................:\n\n");
			 if ( TARGET_HOST != NULL && KEY_FILE != NULL && TARGET_PORT > 0 && LISTEN_PORT > 0 && ENCRYPT_MODE >= 0 && KEY_LEN <= 0 && READ_IDENT >= 0 && SEND_IDENT >= 0 ) exit(2); 
			 show_help = 0;
			break;
		 case 'g':
			 generate_key( argv[2] );
			break;
		}
	}

 if ( show_help ) {
	 printf("\n\t. dBnc .......... (c) 2001 .\n");
	 printf("\t:                          :\n");
	 printf("\t: -h        help           :\n");
	 printf("\t: -r <file> read config    :\n");
	 printf("\t: -g <file> generate key   :\n");
	 printf("\t:                          :\n");
	 printf("\t:..........................:\n\n");
	 exit( 0 );
	}

 bzero( (char *)&local_server_addr, sizeof( local_server_addr ) );

 local_server_addr.sin_addr.s_addr = INADDR_ANY;
 local_server_addr.sin_family = AF_INET;
 local_server_addr.sin_port = htons( (unsigned)LISTEN_PORT );

 server_socket = socket(PF_INET, SOCK_STREAM, 0 );
 server_fd = bind(server_socket, (struct sockaddr *) &local_server_addr, sizeof( local_server_addr ) );

 host_pointer = gethostbyaddr((char *)&local_server_addr.sin_addr, sizeof (struct in_addr), AF_INET);
 if ( host_pointer )
	 strncpy( local_host, host_pointer->h_name, sizeof( local_host ) );
	else
	 strncpy( local_host, inet_ntoa( local_server_addr.sin_addr ), sizeof( local_host ) );

 if ( server_fd < 0 ) {
	 fprintf( stderr, "ERROR: Failed to bind socket!\n");
	 exit(2);
	}

 if ( listen( server_socket, 1024 ) == -1 ) {
	 fprintf( stderr, "ERROR: Could not listed socket!\n");
	 exit(2);
	}

 if ( fork() ) exit( 0 );

 while ( 1 ) {
	 n = sizeof( remote_client_addr );
	 bzero(&remote_client_addr, n);
	 remote_fd = accept( server_socket, (struct sockaddr *)&remote_client_addr, &n );
	 remote_server_addr.sin_family = AF_INET;
	 remote_server_addr.sin_port = htons( (unsigned)TARGET_PORT );
	 host_pointer = gethostbyname( TARGET_HOST[CURRENT_TARGET] );
	 memcpy( &remote_server_addr.sin_addr, host_pointer->h_addr, host_pointer->h_length );

	 switch ( fork() ) {
		 case 0:
			 if ( READ_IDENT == 1 ) {
				 ident = rfc931( &remote_client_addr, &local_server_addr );
				}
			 key_pos = 0;

			 host_pointer = gethostbyaddr((char *)&remote_client_addr.sin_addr, sizeof (struct in_addr), AF_INET);
			 if ( host_pointer )
				 strncpy( remote_client, host_pointer->h_name, sizeof( remote_client ) );
				else
				 strncpy( remote_client, inet_ntoa( remote_client_addr.sin_addr ), sizeof( remote_client ) );

			 close( server_socket );
			 remote_server_socket = socket( AF_INET, SOCK_STREAM, 0 );
			 if ( connect( remote_server_socket, (struct sockaddr *) &remote_server_addr, sizeof( remote_server_addr ) ) == 0 ) {
				 transmit_loop( remote_fd, remote_server_socket, ident, inet_ntoa( remote_client_addr.sin_addr ), remote_client, local_host );
				}
			 CLOSE( remote_server_socket );
			 CLOSE( remote_fd );
			 exit(0);
			break;
		case -1:
			break;
		default:
			 if ( fork() ) exit ( 0 );
			break;
		}
	 close(remote_fd);
	 if ( CURRENT_TARGET < TARGET_HOSTS - 1 ) CURRENT_TARGET++; else CURRENT_TARGET = 0;
	}
}
